package com.outr.webcommunicator

import java.util.UUID
import annotation.tailrec

/**
 * @author Matt Hicks <mhicks@powerscala.org>
 */
trait CommunicatorManager[C <: Communicator] {
  private var map = Map.empty[UUID, C]

  final def create(id: UUID) = {
    val instance = instantiate(id)
    synchronized {
      map += instance.id -> instance
    }
    instance
  }

  final def apply(id: UUID) = map.get(id)

  final def remove(id: UUID) = apply(id) match {
    case Some(c) => synchronized {
      map -= id
      c.dispose()
      true
    }
    case None => false
  }

  def update(): Unit = {
    updateCommunicators()
  }

  @tailrec
  final def updateCommunicators(values: Iterable[C] = this.values): Unit = if (values.nonEmpty) {
    val c = values.head
    c.update()
    if (c.lastUpdated > c.timeout) {    // Hasn't been updated
      remove(c.id)
    }
    updateCommunicators(values.tail)
  }

  def values = map.values

  protected def instantiate(id: UUID): C
}