package com.outr.webcommunicator.proxy

import org.jboss.netty.channel._
import org.jboss.netty.handler.codec.http.{HttpResponseStatus, HttpRequestEncoder, HttpRequest}
import org.jboss.netty.bootstrap.ClientBootstrap
import org.jboss.netty.buffer.ChannelBuffers
import org.powerscala.log.Logging

import com.outr.webcommunicator.netty._
import handler.RequestHandler

/**
 * @author Matt Hicks <mhicks@outr.com>
 */
class ProxyHandler(server: ProxyServer) extends SimpleChannelUpstreamHandler with Logging {
  @volatile var inboundChannel: Channel = _
  @volatile var outboundChannel: Channel = _
  @volatile var outboundFuture: Option[ChannelFuture]= None

  override def channelOpen(ctx: ChannelHandlerContext, e: ChannelStateEvent) {
    inboundChannel = e.getChannel
  }

  override def messageReceived(ctx: ChannelHandlerContext, evt: MessageEvent) = {
    val message = evt.getMessage
    message match {
      case request: HttpRequest if (!outboundFuture.isDefined) => {
        server.findProxy(request) match {
          case Some(address) => {
            info("%s being proxied to %s:%s".format(request2URL(request), address.getHostName, address.getPort))
            val clientBootstrap = new ClientBootstrap(server.clientFactory)
            clientBootstrap.getPipeline.addLast("encoder", new HttpRequestEncoder)
            clientBootstrap.getPipeline.addLast("handler", new OutboundHandler(this))
            outboundFuture = Some(clientBootstrap.connect(address))
            inboundChannel.setReadable(false)

            outboundFuture.get.addListener(new ChannelFutureListener {
              def operationComplete(future: ChannelFuture) = {
                if (future.isSuccess) {
                  outboundChannel = future.getChannel
                  ProxyHandler.this.synchronized {
                    outboundChannel.write(message).addListener(new ChannelFutureListener {
                      def operationComplete(future: ChannelFuture) = {
                        // Remove the decoder and encoder after we finish writing back if upgrading to WebSocket
                        if (request.getHeader("Upgrade") == "websocket") {
                          inboundChannel.getPipeline.remove("decoder")
                          outboundChannel.getPipeline.remove("encoder")
                        }
                      }
                    })
                    // If outboundChannel is saturated, do not read until notified in OutboundHandler.channelInterestChanged
                    inboundChannel.setReadable(outboundChannel.isWritable)
                  }

                } else {
                  inboundChannel.close()              // Connection attempt failed, close the connection
                }
              }
            })
          }
          case None => {                      // No proxy found for URL
            warn("No proxy found for URL: %s".format(request2URL(request)))
            RequestHandler.respond(ctx, HttpResponseStatus.NOT_FOUND)
          }
        }
      }
      case _ => {
        if (outboundChannel == null) {
          synchronized {
            if (outboundChannel == null) {
              outboundFuture.get.addListener(new ChannelFutureListener {
                def operationComplete(future: ChannelFuture) {
                  outboundChannel.write(message)
                  inboundChannel.setReadable(outboundChannel.isWritable)
                }
              })
            } else {
              outboundChannel.write(message)
              inboundChannel.setReadable(outboundChannel.isWritable)
            }
          }
        } else {
          outboundChannel.write(message)
          inboundChannel.setReadable(outboundChannel.isWritable)
        }
      }
    }
  }

  override def channelInterestChanged(ctx: ChannelHandlerContext, e: ChannelStateEvent) {
    // If inboundChannel is not saturated anymore, continue accepting the incoming traffic from the outboundChannel
    synchronized {
      if (e.getChannel.isWritable && outboundChannel != null && outboundChannel.isOpen) {
        inboundChannel.setReadable(true)
      }
    }
  }

  override def channelClosed(ctx: ChannelHandlerContext, e: ChannelStateEvent) {
    if (outboundChannel != null) {
      ProxyHandler.closeOnFlush(outboundChannel)
    }
  }

  override def exceptionCaught(ctx: ChannelHandlerContext, e: ExceptionEvent) {
    e.getCause.printStackTrace()
    ProxyHandler.closeOnFlush(e.getChannel)
  }
}

object ProxyHandler {
  def closeOnFlush(channel: Channel) = {
    if (channel.isConnected) {
      channel.write(ChannelBuffers.EMPTY_BUFFER).addListener(ChannelFutureListener.CLOSE)
    }
  }
}